from typing import Literal

from pydantic import Field, model_validator

from abuse_whois.utils import is_email, is_url

from .api_model import APIModel
from .whois import WhoisRecord

ContactType = Literal["email", "form"]


class Contact(APIModel):
    provider: str = Field(..., description="Provider name")
    address: str = Field(..., description="Contact address")
    type: ContactType = Field(default="email", description="Type of contact method")

    @model_validator(mode="after")
    def change_type(self):
        if is_email(self.address):
            self.type = "email"
        elif is_url(self.address):
            self.type = "form"

        return self


class ContactsRecords(APIModel):
    domain: WhoisRecord | None = Field(
        default=None, description="Domain Whois/RDAP record"
    )
    ip_address: WhoisRecord | None = Field(
        default=None, description="IP Address Whois/RDAP record"
    )


class Contacts(APIModel):
    address: str
    hostname: str = Field(..., description="Hostname")

    ip_address: str | None = Field(None, description="IP address")
    registered_domain: str | None = Field(
        None, description="Registered domain (a.k.a. free level domain)"
    )

    shared_hosting_provider: Contact | None = Field(
        None, description="Shared hosting provider"
    )
    registrar: Contact | None = Field(None, description="Registrar")
    hosting_provider: Contact | None = Field(None, description="Hosting provider")

    records: ContactsRecords = Field(..., description="Domain/IP Whois records")



import pytest

from abuse_whois import schemas


@pytest.mark.parametrize(
    "address,expected",
    [
        ("http://example.com", "form"),
        ("abuse@example.com", "email"),
    ],
)
def test_contact_type(address: str, expected: str):
    assert schemas.Contact(provider="foo", address=address).type == expected



import contextlib
from functools import partial
from urllib.parse import urlparse

import aiometer
from pydantic import BaseModel
from returns.functions import raise_exception
from returns.future import FutureResultE, future_safe
from returns.pipeline import flow
from returns.pointfree import bind
from returns.result import ResultE, safe
from returns.unsafe import unsafe_perform_io

from abuse_whois import errors, schemas, settings
from abuse_whois.matchers.shared_hosting import get_shared_hosting_provider
from abuse_whois.matchers.whois import get_whois_contact
from abuse_whois.utils import (
    get_registered_domain,
    is_domain,
    is_email,
    is_ipv4,
    is_ipv6,
    is_url,
    resolve,
)
from abuse_whois.whois import query

from .abstract import AbstractService


class Container(BaseModel):
    hostname: str
    ip_address: str | None
    domain_record: schemas.WhoisRecord | None
    ip_record: schemas.WhoisRecord | None


@future_safe
async def validate_address(address: str) -> str:
    for f in [is_email, is_domain, is_ipv4, is_ipv6, is_url]:
        if f(address):
            return address

    raise errors.AddressError(f"Address:{address} not supported address type")


@future_safe
async def get_hostname(address: str) -> str:
    if is_ipv6(address) or is_ipv4(address) or is_domain(address):
        return address

    url_or_email = address
    if is_email(url_or_email):
        url_or_email = f"http://{address}"

    parsed = urlparse(url_or_email)
    if parsed.hostname is None:
        raise errors.AddressError(f"Address:{address} does not have hostname")

    return parsed.hostname


@future_safe
async def safe_query(
    hostname, *, timeout: int = settings.QUERY_TIMEOUT
) -> schemas.WhoisRecord:
    return await query(hostname, timeout=timeout)


async def get_optional_ip_record(
    hostname: str | None, *, timeout: int = settings.QUERY_TIMEOUT
) -> schemas.WhoisRecord | None:
    if hostname is None:
        return None

    ip_f_result = await safe_query(hostname, timeout=timeout)
    return unsafe_perform_io(ip_f_result.value_or(None))


async def get_optional_domain_record(
    hostname: str, *, timeout: int = settings.QUERY_TIMEOUT
) -> schemas.WhoisRecord | None:
    if not is_domain(hostname):
        return None

    domain_f_result = await safe_query(hostname, timeout=timeout)
    return unsafe_perform_io(domain_f_result.value_or(None))


@future_safe
async def get_records(
    hostname: str, *, timeout: int = settings.QUERY_TIMEOUT
) -> Container:
    domain_record: schemas.WhoisRecord | None = None
    ip_record: schemas.WhoisRecord | None = None

    ip_address: str | None = None
    if is_domain(hostname):
        # get IP address by domain
        with contextlib.suppress(OSError):
            ip_address = await resolve(hostname)
    else:
        ip_address = hostname

    tasks = [
        partial(get_optional_domain_record, hostname, timeout=timeout),
        partial(get_optional_ip_record, ip_address, timeout=timeout),
    ]
    domain_record, ip_record = await aiometer.run_all(tasks)
    return Container(
        hostname=hostname,
        domain_record=domain_record,
        ip_record=ip_record,
        ip_address=ip_address,
    )


@future_safe
async def get_contacts(container: Container, *, address: str) -> schemas.Contacts:
    @safe
    def set_records(container: Container) -> schemas.Contacts:
        if container.ip_record is None and container.domain_record is None:
            raise errors.NotFoundError(
                f"Record: {container.hostname} not found or something went wrong"
            )

        return schemas.Contacts(
            records=schemas.ContactsRecords(
                ip_address=container.ip_record, domain=container.domain_record
            ),
            address=address,
            ip_address=container.ip_address,
            hostname=container.hostname,
            registered_domain=None,
            shared_hosting_provider=None,
            hosting_provider=None,
            registrar=None,
        )

    @safe
    def set_registered_domain(contacts: schemas.Contacts) -> schemas.Contacts:
        if is_domain(container.hostname):
            contacts.registered_domain = get_registered_domain(container.hostname)

        return contacts

    @safe
    def set_registrar(contacts: schemas.Contacts) -> schemas.Contacts:
        if container.domain_record is not None:
            contacts.registrar = get_whois_contact(container.domain_record)

        return contacts

    @safe
    def set_hosting_provider(contacts: schemas.Contacts) -> schemas.Contacts:
        if container.ip_record is not None:
            contacts.hosting_provider = get_whois_contact(container.ip_record)

        return contacts

    @safe
    def set_shared_hosting_provider(
        contacts: schemas.Contacts,
    ) -> schemas.Contacts:
        contacts.shared_hosting_provider = get_shared_hosting_provider(
            container.hostname
        )
        return contacts

    result: ResultE[schemas.Contacts] = flow(
        set_records(container),
        bind(set_registered_domain),
        bind(set_registrar),
        bind(set_hosting_provider),
        bind(set_shared_hosting_provider),
    )
    return result.alt(raise_exception).unwrap()


class ContactsQuery(AbstractService):
    async def call(self, address: str) -> schemas.Contacts:
        f_result: FutureResultE[schemas.Contacts] = flow(
            address,
            validate_address,
            bind(get_hostname),
            bind(get_records),
            bind(partial(get_contacts, address=address)),
        )
        result = await f_result.awaitable()
        return unsafe_perform_io(result.alt(raise_exception).unwrap())



import asyncio
import json
import sys

import typer
from returns.result import Failure, Success, safe

from . import schemas, services

app = typer.Typer()


async def get_abuse_contacts(address: str) -> schemas.Contacts:
    return await services.ContactsQuery().call(address)


@app.command()
def whois(
    address: str = typer.Argument(..., help="URL, domain, IP address or email address"),
):
    @safe
    def inner() -> schemas.Contacts:
        return asyncio.run(get_abuse_contacts(address))

    result = inner()
    match result:
        case Success(value):
            print(value.model_dump_json(by_alias=True))  # noqa: T201
        case Failure(e):
            print(json.dumps({"detail": str(e)}), file=sys.stderr)  # noqa: T201
            sys.exit(1)


if __name__ == "__main__":
    app()



import asyncio

import pytest
from pytest_mock import MockerFixture

from abuse_whois.utils import is_domain, is_email, is_ipv4, is_url, resolve


@pytest.mark.parametrize(
    "v,expected",
    [
        ("1.1.1.1", True),
        ("example.com", False),
        ("https://github.com", False),
        ("foo@test.com", False),
    ],
)
def test_is_ipv4(v: str, expected: bool):
    assert is_ipv4(v) is expected


@pytest.mark.parametrize(
    "v,expected",
    [
        ("1.1.1.1", False),
        ("example.com", True),
        ("https://github.com", False),
        ("foo@test.com", False),
    ],
)
def test_is_domain(v: str, expected: bool):
    assert is_domain(v) is expected


@pytest.mark.parametrize(
    "v,expected",
    [
        ("1.1.1.1", False),
        ("example.com", False),
        ("https://github.com", True),
        ("foo@test.com", False),
    ],
)
def test_is_url(v: str, expected: bool):
    assert is_url(v) is expected


@pytest.mark.parametrize(
    "v,expected",
    [
        ("1.1.1.1", False),
        ("example.com", False),
        ("https://github.com", False),
        ("foo@test.com", True),
    ],
)
def test_is_email(v: str, expected: bool):
    assert is_email(v) is expected


@pytest.mark.asyncio
async def test_resolve_with_timeout(mocker: MockerFixture):
    with pytest.raises(asyncio.TimeoutError):
        assert await resolve("github.com", timeout=-1.0)



import azuma

from .api_model import APIModel, RootAPIModel
from .contact import Contact


class Rule(APIModel, azuma.Rule):
    contact: Contact


class RuleSet(RootAPIModel, azuma.RuleSet):
    pass



import glob
import json

import pytest

from abuse_whois.matchers.whois import WhoisRule, WhoisRuleSet
from abuse_whois.schemas import WhoisRecord

paths = list(glob.glob("abuse_whois/matchers/whois/rules/*.yaml"))


@pytest.mark.parametrize("path", paths)
def test_load_rules(path: str):
    assert WhoisRule.model_validate_file(path)


@pytest.fixture
def godaddy_whois_record():
    with open("tests/fixtures/godaddy.com.json") as f:
        data = json.loads(f.read())
        return WhoisRecord.model_validate(data)


@pytest.fixture
def godaddy_whois_rule():
    return WhoisRule.model_validate_file(
        "abuse_whois/matchers/whois/rules/godaddy.yaml"
    )


def test_godaddy(godaddy_whois_record: WhoisRecord, godaddy_whois_rule: WhoisRule):
    condition = godaddy_whois_rule.detection.condition
    assert (
        condition(godaddy_whois_rule, godaddy_whois_record.model_dump(by_alias=True))
        is True
    )
    assert condition(godaddy_whois_rule, {}) is False


def test_rule_set():
    rule_set = WhoisRuleSet.from_dir()
    assert len(rule_set.root) > 0



from pydantic import BaseModel, ConfigDict, RootModel
from pydantic.alias_generators import to_camel


class APIModel(BaseModel):
    model_config = ConfigDict(
        from_attributes=True, alias_generator=to_camel, populate_by_name=True
    )


class RootAPIModel(RootModel):
    def __iter__(self):
        return iter(self.root)

    def __getitem__(self, item):
        return self.root[item]

    model_config = ConfigDict(
        from_attributes=True, alias_generator=to_camel, populate_by_name=True
    )



import asyncio

from fastapi import APIRouter, HTTPException, status
from returns.future import future_safe
from returns.pipeline import is_successful
from returns.unsafe import unsafe_perform_io

from abuse_whois import errors, get_abuse_contacts, schemas

router = APIRouter(prefix="/whois")


@router.post("/", response_model=schemas.Contacts)
async def whois(query: schemas.Query) -> schemas.Contacts:
    @future_safe
    async def inner():
        return await get_abuse_contacts(query.address)

    result = await inner()
    if not is_successful(result):
        failure = unsafe_perform_io(result.failure())
        status_code = status.HTTP_400_BAD_REQUEST

        match failure:
            case errors.NotFoundError():
                status_code = status.HTTP_404_NOT_FOUND
            case asyncio.TimeoutError():
                status_code = status.HTTP_408_REQUEST_TIMEOUT
            case errors.RateLimitError():
                status_code = status.HTTP_429_TOO_MANY_REQUESTS
            case errors.AbuseWhoisError():
                status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
            case ConnectionResetError():
                status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

        raise HTTPException(status_code, detail=str(failure))

    return unsafe_perform_io(result.unwrap())



import sys

from starlette.config import Config

config = Config(".env")

# FastAPI settings
PROJECT_NAME: str = config("PROJECT_NAME", default="abuse-whois")

DEBUG: bool = config("DEBUG", cast=bool, default=False)
TESTING: bool = config("TESTING", cast=bool, default=False)

LOG_FILE = config("LOG_FILE", default=sys.stderr)
LOG_LEVEL: str = config("LOG_LEVEL", cast=str, default="DEBUG")
LOG_BACKTRACE: bool = config("LOG_BACKTRACE", cast=bool, default=True)

# Query settings
QUERY_TIMEOUT: int = config("QUERY_TIMEOUT", cast=int, default=10)
QUERY_CACHE_SIZE: int = config("QUERY_CACHE_SIZE", cast=int, default=1024)
QUERY_CACHE_TTL: int = config("QUERY_CACHE_TTL", cast=int, default=60 * 60)
QUERY_MAX_RETRIES: int = config("QUERY_MAX_RETRIES", cast=int, default=3)

# Rule settings
ADDITIONAL_WHOIS_RULE_DIRECTORY: str | None = config(
    "ADDITIONAL_WHOIS_RULE_DIRECTORY", cast=str, default=None
)
ADDITIONAL_SHARED_HOSTING_RULE_DIRECTORY: str | None = config(
    "ADDITIONAL_SHARED_HOSTING_RULE_DIRECTORY", cast=str, default=None
)

# Etc.
WHOIS_RATE_LIMIT_MESSAGES: set[str] = {
    "WHOIS LIMIT EXCEEDED - SEE WWW.PIR.ORG/WHOIS FOR DETAILS",
    "Your access is too fast,please try again later.",
    "Your connection limit exceeded.",
    "Number of allowed queries exceeded.",
    "WHOIS LIMIT EXCEEDED",
    "Requests of this client are not permitted.",
    "Too many connection attempts. Please try again in a few seconds.",
    "We are unable to process your request at this time.",
    "HTTP/1.1 400 Bad Request",
    "Closing connections because of Timeout",
    "Access to whois service at whois.isoc.org.il was **DENIED**",
    "IP Address Has Reached Rate Limit",
}



import os

max_workers_str = os.getenv("MAX_WORKERS", "2")
use_max_workers = int(max_workers_str)

host = os.getenv("HOST", "0.0.0.0")
port = os.getenv("PORT", "8000")
use_loglevel = os.getenv("LOG_LEVEL", "info")
use_bind = f"{host}:{port}"

accesslog_var = os.getenv("ACCESS_LOG", "-")
use_accesslog = accesslog_var or None
errorlog_var = os.getenv("ERROR_LOG", "-")
use_errorlog = errorlog_var or None
graceful_timeout_str = os.getenv("GRACEFUL_TIMEOUT", "120")
timeout_str = os.getenv("TIMEOUT", "120")
keepalive_str = os.getenv("KEEP_ALIVE", "5")


# Gunicorn config variables
loglevel = use_loglevel
workers = use_max_workers
bind = use_bind
errorlog = use_errorlog
accesslog = use_accesslog
graceful_timeout = int(graceful_timeout_str)
timeout = int(timeout_str)
keepalive = int(keepalive_str)



import pytest

from abuse_whois.main import get_abuse_contacts


@pytest.mark.parametrize(
    "address,hostname",
    [
        ("1.1.1.1", "1.1.1.1"),
        ("example.com", "example.com"),
        ("https://github.com", "github.com"),
        ("foo@test.com", "test.com"),
    ],
)
@pytest.mark.asyncio
async def test_get_abuse_contacts(address: str, hostname: str):
    contacts = await get_abuse_contacts(address)
    assert contacts.hostname == hostname


@pytest.mark.asyncio
async def test_get_abuse_contacts_with_ip_address():
    contacts = await get_abuse_contacts("1.1.1.1")
    assert contacts.hostname == "1.1.1.1"
    assert contacts.ip_address == "1.1.1.1"
    assert contacts.registered_domain is None
    assert contacts.registrar is None
    assert contacts.hosting_provider is not None


@pytest.mark.asyncio
async def test_get_abuse_contacts_with_domain():
    contacts = await get_abuse_contacts("www.github.com")
    assert contacts.hostname == "www.github.com"
    assert contacts.ip_address is not None
    assert contacts.registered_domain == "github.com"
    assert contacts.registrar is not None



import typing
from abc import ABC, abstractmethod


class AbstractService(ABC):
    @abstractmethod
    async def call(self, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        raise NotImplementedError()



import glob

import pytest

from abuse_whois.matchers.shared_hosting import (
    SharedHostingRule,
    SharedHostingRuleSet,
    get_shared_hosting_provider,
)

paths = list(glob.glob("abuse_whois/matchers/shared_hosting/rules/*.yaml"))


@pytest.mark.parametrize("path", paths)
def test_load_rules(path: str):
    assert SharedHostingRule.model_validate_file(path)


@pytest.fixture
def bitly_rule():
    return SharedHostingRule.model_validate_file(
        "abuse_whois/matchers/shared_hosting/rules/bitly.yaml"
    )


def test_bit_ly(bitly_rule: SharedHostingRule):
    condition = bitly_rule.detection.condition
    assert condition(bitly_rule, {"domain": "bit.ly"}) is True
    assert condition(bitly_rule, {"domain": ""}) is False


@pytest.mark.parametrize(
    "hostname",
    ["foo.bit.ly", "foo.blogger.com"],
)
def test_get_shared_hosting_provider(hostname: str):
    assert get_shared_hosting_provider(hostname) is not None


def test_shared_hosting_rule_set():
    rule_set = SharedHostingRuleSet.from_dir()
    assert len(rule_set.root) > 0



import asyncio
import socket
from collections.abc import Callable, Iterable
from contextlib import contextmanager
from typing import Any, TypeVar

import tld
import validators
from asyncer import asyncify

from . import settings


def _is_x(v: str, *, validator: Callable[[str], bool]) -> bool:
    res = validator(v)

    if isinstance(res, validators.ValidationError):
        return False

    return res


def is_ipv4(v: str) -> bool:
    return _is_x(v, validator=validators.ipv4)  # type: ignore


def is_ipv6(v: str) -> bool:
    return _is_x(v, validator=validators.ipv6)  # type: ignore


def is_domain(v: str) -> bool:
    if len(v.split(".")) == 1:
        return False

    if "@" in v:
        return False

    return _is_x(v, validator=validators.domain)  # type: ignore


def is_url(v: str) -> bool:
    if not v.startswith(("http://", "https://")):
        return False

    return _is_x(v, validator=validators.url)  # type: ignore


def is_email(v: str) -> bool:
    return _is_x(v, validator=validators.email)  # type: ignore


def get_registered_domain(v: str) -> str | None:
    registered_domain = tld.get_fld(v, fail_silently=True, fix_protocol=True)

    if registered_domain == "":
        return None

    return registered_domain


@contextmanager
def with_socket_timeout(timeout: float):
    old = socket.getdefaulttimeout()
    try:
        socket.setdefaulttimeout(timeout)
        yield
    except (TimeoutError, ValueError) as e:
        raise asyncio.TimeoutError(
            f"{timeout} seconds have passed but there is no response"
        ) from e
    finally:
        socket.setdefaulttimeout(old)


def _resolve(hostname: str, *, timeout: float = float(settings.QUERY_TIMEOUT)) -> str:
    with with_socket_timeout(timeout):
        return socket.gethostbyname(hostname)


resolve = asyncify(_resolve)


def is_iterable(obj: Any) -> bool:
    try:
        iter(obj)
    except TypeError:
        return False
    return True


T = TypeVar("T")


def unique(src: Iterable[T], key: str | Callable[[T], Any] | None = None) -> list[T]:
    return list(unique_iter(src, key))


def unique_iter(src: Iterable[T], key: str | Callable[[T], Any] | None = None):
    if not is_iterable(src):
        raise TypeError(f"expected an iterable, not {type(src)!r}")

    def build_key_func():
        if key is None:
            return lambda x: x

        if callable(key):
            return key

        if isinstance(key, str):
            return lambda x: getattr(x, key, x)

        raise TypeError(f'"key" expected a string or callable, not {key!r}')

    key_func = build_key_func()
    seen = set()
    for i in src:
        k = key_func(i)
        if k not in seen:
            seen.add(k)
            yield i



from pathlib import Path

import senkawa
from returns.maybe import Maybe

from abuse_whois import schemas, settings
from abuse_whois.utils import is_email, unique

DEFAULT_RULE_DIRECTORY: Path = Path(__file__).parent / "./rules"


class WhoisRule(schemas.Rule):
    def match(self, record: schemas.WhoisRecord) -> bool:
        return super().match(record.model_dump(by_alias=True))


class WhoisRuleSet(schemas.RootAPIModel):
    root: list[WhoisRule]

    @classmethod
    def from_dir(cls, dir: str | Path = DEFAULT_RULE_DIRECTORY):
        dir = Path(dir) if isinstance(dir, str) else dir
        expanded = senkawa.glob(str(dir.joinpath("*.{yaml,yml}")))
        return cls(root=[WhoisRule.model_validate_file(p) for p in expanded])


def load_rule_set():
    base = WhoisRuleSet.from_dir(DEFAULT_RULE_DIRECTORY)
    addition = (
        WhoisRuleSet(root=[])
        if settings.ADDITIONAL_WHOIS_RULE_DIRECTORY is None
        else WhoisRuleSet.from_dir(settings.ADDITIONAL_WHOIS_RULE_DIRECTORY)
    )
    return WhoisRuleSet(root=unique(addition.root + base.root, "id"))


default_rule_set = load_rule_set()


def normalize_abuse_email(email: str | None) -> str | None:
    return (
        Maybe.from_optional(email)
        .bind_optional(lambda x: x.removeprefix("mailto:"))
        .value_or(None)
    )


def get_provider(record: schemas.WhoisRecord) -> str | None:
    def inner(email: str) -> str | None:
        if not is_email(email):
            return None

        # use email's domain as a provider if provider is None
        return email.split("@")[-1]

    return (
        Maybe.from_optional(normalize_abuse_email(record.abuse.email))
        .bind_optional(inner)
        .value_or(record.registrar)
    )


def get_address(record: schemas.WhoisRecord) -> str | None:
    def inner(email: str) -> str | None:
        # check email format for just in case
        if not is_email(email):
            return None

        return email

    return (
        Maybe.from_optional(normalize_abuse_email(record.abuse.email))
        .bind_optional(inner)
        .value_or(None)
    )


def get_registrar_contact(record: schemas.WhoisRecord) -> schemas.Contact | None:
    provider = get_provider(record)
    address = get_address(record)

    if provider is None or address is None:
        return None

    return schemas.Contact(provider=provider, address=address)


def get_whois_contact(
    record: schemas.WhoisRecord, *, rules: list[WhoisRule] | None = None
) -> schemas.Contact | None:
    rule_set = WhoisRuleSet(root=rules or default_rule_set.root)  # type: ignore
    for rule in rule_set:
        if rule.match(record):
            return rule.contact

    return get_registrar_contact(record)



from functools import partial
from typing import Any, cast

from asyncwhois.client import DomainClient, NumberClient
from asyncwhois.errors import NotFoundError as WhoIsNotFoundError
from asyncwhois.parse import IPBaseKeys, TLDBaseKeys
from returns.functions import raise_exception
from returns.future import FutureResultE, future_safe
from returns.maybe import Maybe
from returns.pipeline import flow, is_successful
from returns.pointfree import bind
from returns.unsafe import unsafe_perform_io
from whodap import DomainResponse
from whodap.errors import NotFoundError as WhodapNotFoundError
from whodap.errors import RateLimitError

from abuse_whois import errors, schemas, settings
from abuse_whois.utils import (
    get_registered_domain,
    is_domain,
    is_ipv4,
    is_ipv6,
)

from .abstract import AbstractService

QueryOutput = (
    tuple[str, dict[str, Any]]
    | tuple[str, dict[TLDBaseKeys, Any]]
    | tuple[str, dict[IPBaseKeys, Any]]
)


@future_safe
async def domain_rdap(domain: str, *, client: DomainClient) -> QueryOutput:
    return await client.aio_rdap(domain)


@future_safe
async def domain_whois(domain: str, *, client: DomainClient) -> QueryOutput:
    return await client.aio_whois(domain)


@future_safe
async def raise_on_rate_limit(query_output: QueryOutput) -> QueryOutput:
    query_string, _ = query_output
    for message in settings.WHOIS_RATE_LIMIT_MESSAGES:
        if message in query_string:
            raise RateLimitError()

    return query_output


@future_safe
async def rdap(hostname: str, *, client: DomainClient | NumberClient):
    return await client.aio_rdap(hostname)


@future_safe
async def whois(hostname: str, *, client: DomainClient | NumberClient):
    return await client.aio_whois(hostname)


async def _query(hostname: str, *, client: DomainClient | NumberClient):
    rdap_result = await rdap(hostname, client=client)
    if is_successful(rdap_result):
        query_string, parsed = unsafe_perform_io(rdap_result.unwrap())

        registrar = parsed.get("registrar")
        email = parsed.get("registrar_abuse_email")
        if registrar is not None or email is not None:
            return query_string, parsed

    whois_f_result: FutureResultE[QueryOutput] = flow(
        whois(hostname, client=client), bind(raise_on_rate_limit)
    )
    whois_result = await whois_f_result.awaitable()
    return unsafe_perform_io(whois_result.alt(raise_exception).unwrap())


@future_safe
async def query(address: str, *, timeout: int = settings.QUERY_TIMEOUT) -> QueryOutput:
    @future_safe
    async def inner() -> QueryOutput:
        if is_domain(address):
            return await _query(address, client=DomainClient(timeout=timeout))

        if is_ipv4(address) or is_ipv6(address):
            return await _query(address, client=NumberClient(timeout=timeout))

        raise errors.AddressError(f"{address} is neither of domain, IPv4 and IPv6")

    f_result = inner()
    result = await f_result.awaitable()
    if not is_successful(result):
        failure = unsafe_perform_io(result.failure())
        match failure:
            case WhodapNotFoundError() | WhoIsNotFoundError():
                raise errors.NotFoundError(f"Record:{address} not found") from failure
            case RateLimitError():
                raise errors.RateLimitError(
                    f"Query:{address} rate limited"
                ) from failure

    return unsafe_perform_io(result.alt(raise_exception).unwrap())


def get_contact(parsed: dict, prefix: str) -> schemas.WhoisContact:
    return schemas.WhoisContact(
        name=parsed.get(f"{prefix}_name", None),
        email=parsed.get(f"{prefix}_email", None),
        telephone=parsed.get(f"{prefix}_phone", None),
        organization=parsed.get(f"{prefix}_organization", None),
    )


def get_abuse(parsed: dict) -> schemas.WhoisAbuse:
    return schemas.WhoisAbuse(
        email=parsed.get("registrar_abuse_email"),
        telephone=parsed.get("registrar_abuse_phone"),
    )


@future_safe
async def normalize(hostname: str) -> str:
    if is_domain(hostname):
        return get_registered_domain(hostname) or hostname

    return hostname


def normalize_domain(domain: DomainResponse | str | None) -> str | None:
    if domain is None:
        return None

    if isinstance(domain, str):
        return domain.lower().removesuffix(".")

    return (
        Maybe.from_optional(domain.to_dict().get("stringValue"))
        .bind_optional(lambda x: str(x).lower().removesuffix("."))
        .value_or(None)
    )


def is_str_list(values: Any) -> bool:
    if not isinstance(values, list):
        return False

    return all(isinstance(v, str) for v in values)


def get_str_list(data: dict[str, Any], key: str) -> list[str]:
    return cast(
        list[str],
        (
            Maybe.from_optional(data.get(key))
            .bind_optional(lambda x: x if is_str_list(x) else [])
            .value_or([])
        ),
    )


@future_safe
async def parse(query_output: QueryOutput) -> schemas.WhoisRecord:
    query_string, parsed = query_output
    parsed = cast(dict[str, Any], parsed)
    return schemas.WhoisRecord(
        raw_text=query_string,
        domain=normalize_domain(parsed.get("domain_name")),
        name_servers=get_str_list(parsed, "name_servers"),
        statuses=get_str_list(parsed, "status"),
        tech=get_contact(parsed, "technical"),
        admin=get_contact(parsed, "admin"),
        registrant=get_contact(parsed, "registrant"),
        abuse=get_abuse(parsed),
        expires_at=parsed.get("expires"),
        updated_at=parsed.get("updated"),
        registered_at=parsed.get("registered"),
        registrar=parsed.get("registrar"),
    )


class WhoisQuery(AbstractService):
    async def call(
        self, hostname: str, *, timeout: int = settings.QUERY_TIMEOUT
    ) -> schemas.WhoisRecord:
        f_result: FutureResultE[schemas.WhoisRecord] = flow(
            hostname, normalize, bind(partial(query, timeout=timeout)), bind(parse)
        )
        result = await f_result.awaitable()
        return unsafe_perform_io(result.alt(raise_exception).unwrap())



from pathlib import Path

import senkawa

from abuse_whois import schemas, settings
from abuse_whois.utils import unique

DEFAULT_RULE_DIRECTORY: Path = Path(__file__).parent / "./rules"


class SharedHostingRule(schemas.Rule):
    def match(self, hostname: str) -> bool:
        return super().match({"domain": hostname})


class SharedHostingRuleSet(schemas.RootAPIModel):
    root: list[SharedHostingRule]

    @classmethod
    def from_dir(cls, dir: str | Path = DEFAULT_RULE_DIRECTORY):
        dir = Path(dir) if isinstance(dir, str) else dir
        expanded = senkawa.glob(str(dir.joinpath("*.{yaml,yml}")))
        return cls(root=[SharedHostingRule.model_validate_file(p) for p in expanded])


def load_rule_set():
    base = SharedHostingRuleSet.from_dir()
    addition = (
        SharedHostingRuleSet(root=[])
        if settings.ADDITIONAL_WHOIS_RULE_DIRECTORY is None
        else SharedHostingRuleSet.from_dir(settings.ADDITIONAL_WHOIS_RULE_DIRECTORY)
    )
    return SharedHostingRuleSet(root=unique(addition.root + base.root, key="id"))


default_rule_set = load_rule_set()


def get_shared_hosting_provider(
    hostname: str, *, rules: list[SharedHostingRule] | None = None
) -> schemas.Contact | None:
    rule_set = SharedHostingRuleSet(root=rules or default_rule_set.root)
    for rule in rule_set:
        if rule.match(hostname):
            return rule.contact

    return None



from .api_model import APIModel, RootAPIModel  # noqa: F401
from .contact import Contact, Contacts, ContactsRecords  # noqa: F401
from .query import Query  # noqa: F401
from .rule import Rule, RuleSet  # noqa: F401
from .whois import WhoisAbuse, WhoisContact, WhoisRecord  # noqa: F401



import asyncio

import stamina
from asyncache import cached
from cachetools import TTLCache

from . import errors, services, settings


@stamina.retry(
    on=(asyncio.TimeoutError, errors.RateLimitError),
    attempts=settings.QUERY_MAX_RETRIES,
    timeout=None,
)
@cached(cache=TTLCache(maxsize=settings.QUERY_CACHE_SIZE, ttl=settings.QUERY_CACHE_TTL))
async def query(
    hostname: str,
    *,
    timeout: int = settings.QUERY_TIMEOUT,
):
    return await services.WhoisQuery().call(hostname=hostname, timeout=timeout)
